@@ -84,6 +84,10 @@ struct ScalarHashImpl {
     return StdHash(s.value.days) & StdHash(s.value.months) & StdHash(s.value.nanoseconds);
   }
 
+  Status Visit(const Decimal32Scalar& s) { return StdHash(s.value.value()); }
+
+  Status Visit(const Decimal64Scalar& s) { return StdHash(s.value.value()); }
+
   Status Visit(const Decimal128Scalar& s) {
     return StdHash(s.value.low_bits()) & StdHash(s.value.high_bits());
   }
@@ -290,6 +294,24 @@ struct ScalarValidateImpl {
     return Status::OK();
   }
 
+  Status Visit(const Decimal32Scalar& s) {
+    const auto& ty = checked_cast<const DecimalType&>(*s.type);
+    if (!s.value.FitsInPrecision(ty.precision())) {
+      return Status::Invalid("Decimal value ", s.value.ToIntegerString(),
+                             " does not fit in precision of ", ty);
+    }
+    return Status::OK();
+  }
+
+  Status Visit(const Decimal64Scalar& s) {
+    const auto& ty = checked_cast<const DecimalType&>(*s.type);
+    if (!s.value.FitsInPrecision(ty.precision())) {
+      return Status::Invalid("Decimal value ", s.value.ToIntegerString(),
+                             " does not fit in precision of ", ty);
+    }
+    return Status::OK();
+  }
+
   Status Visit(const Decimal128Scalar& s) {
     const auto& ty = checked_cast<const DecimalType&>(*s.type);
     if (!s.value.FitsInPrecision(ty.precision())) {
@@ -1223,7 +1245,7 @@ Result<std::shared_ptr<Scalar>> CastImpl(const StringScalar& from,
   ARROW_ASSIGN_OR_RAISE(auto out,
                         Scalar::Parse(std::move(to_type), std::string_view(*from.value)));
   DCHECK(checked_pointer_cast<ToScalar>(out) != nullptr);
-  return std::move(out);
+  return out;
 }
 
 // binary/large binary/large string to string
@@ -1347,7 +1369,7 @@ struct FromTypeVisitor : CastImplVisitor {
     ARROW_ASSIGN_OR_RAISE(
         out_, CastImpl<ToType>(
                   checked_cast<const typename TypeTraits<FromType>::ScalarType&>(from_),
-                  std::move(to_type_)));
+                  to_type_));
     return Status::OK();
   }
 
@@ -1355,8 +1377,8 @@ struct FromTypeVisitor : CastImplVisitor {
   template <typename T1 = ToType>
   typename std::enable_if_t<TypeTraits<T1>::is_parameter_free, Status> Visit(
       const ToType&) {
-    ARROW_ASSIGN_OR_RAISE(out_, MakeScalar(std::move(to_type_),
-                                           checked_cast<const ToScalar&>(from_).value));
+    ARROW_ASSIGN_OR_RAISE(
+        out_, MakeScalar(to_type_, checked_cast<const ToScalar&>(from_).value));
     return Status::OK();
   }
 
