@@ -15,7 +15,7 @@
 # specific language governing permissions and limitations
 # under the License.
 
-#### Fuctions #### check end of file for main logic
+#### Functions #### check end of file for main logic
 env_is <- function(var, value) identical(tolower(Sys.getenv(var)), value)
 
 # Log messages in the style of the configure script
@@ -112,9 +112,6 @@ validate_checksum <- function(binary_url, libfile, hush = quietly) {
     # Munge the path to the correct sha file which we include during the
     # release process
     checksum_file <- sub(".+/bin/(.+\\.zip)", "\\1\\.sha512", binary_url)
-    if (grepl("rwinlib", binary_url)) {
-      checksum_file <- "windows/arrow-13.0.0.zip.sha512"
-    }
     checksum_file <- file.path(checksum_path, checksum_file)
 
     # Try `shasum`, and if that doesn't work, fall back to `sha512sum` if not found
@@ -155,12 +152,6 @@ validate_checksum <- function(binary_url, libfile, hush = quietly) {
 download_binary <- function(lib) {
   libfile <- paste0("arrow-", VERSION, ".zip")
   binary_url <- paste0(arrow_repo, "bin/", lib, "/arrow-", VERSION, ".zip")
-
-  if (lib == "windows") {
-    binary_url <- "https://github.com/rwinlib/arrow/archive/refs/tags/v13.0.0.zip"
-    libfile <- "arrow-13.0.0.zip"
-  }
-
   if (try_download(binary_url, libfile) && validate_checksum(binary_url, libfile)) {
     lg("Successfully retrieved libarrow (%s)", lib)
   } else {
@@ -230,8 +221,8 @@ check_allowlist <- function(os, allowed = "https://raw.githubusercontent.com/apa
   allowlist <- tryCatch(
     # Try a remote allowlist so that we can add/remove without a release
     suppressWarnings(readLines(allowed)),
-    # Fallback to default: allowed only on Ubuntu and CentOS/RHEL
-    error = function(e) c("ubuntu", "centos", "redhat", "rhel")
+    # Fallback to default allow list shipped with the package
+    error = function(e) readLines("tools/nixlibs-allowlist.txt")
   )
   # allowlist should contain valid regular expressions (plain strings ok too)
   any(grepl(paste(allowlist, collapse = "|"), os))
@@ -395,9 +386,7 @@ distro <- function() {
   out$id <- tolower(out$id)
   # debian unstable & testing lsb_release `version` don't include numbers but we can map from pretty name
   if (is.null(out$version) || out$version %in% c("testing", "unstable")) {
-    if (grepl("bullseye", out$codename)) {
-      out$short_version <- "11"
-    } else if (grepl("bookworm", out$codename)) {
+    if (grepl("bookworm", out$codename)) {
       out$short_version <- "12"
     }
   } else if (out$id == "ubuntu") {
@@ -547,7 +536,7 @@ build_libarrow <- function(src_dir, dst_dir) {
   }
   cleanup(build_dir)
 
-  env_var_list <- c(
+  env_var_list <- list(
     SOURCE_DIR = src_dir,
     BUILD_DIR = build_dir,
     DEST_DIR = dst_dir,
@@ -585,6 +574,14 @@ build_libarrow <- function(src_dir, dst_dir) {
         env_var_list <- c(env_var_list, setNames("BUNDLED", env_var))
       }
     }
+    # We also _do_ want to enable S3 and ZSTD by default
+    # so that binaries built on CRAN from source are fully featured
+    # but defer to the env vars if those are set
+    env_var_list <- c(
+      env_var_list,
+      ARROW_S3 = Sys.getenv("ARROW_S3", "ON"),
+      ARROW_WITH_ZSTD = Sys.getenv("ARROW_WITH_ZSTD", "ON")
+    )
   }
 
   env_var_list <- with_cloud_support(env_var_list)
@@ -825,8 +822,16 @@ set_thirdparty_urls <- function(env_var_list) {
   env_var_list
 }
 
-is_feature_requested <- function(env_varname, default = env_is("LIBARROW_MINIMAL", "false")) {
-  env_value <- tolower(Sys.getenv(env_varname))
+# this is generally about features that people asked for via environment variables, but
+# for some cases (like S3 when we override it in this script) we might find those in
+# env_var_list
+is_feature_requested <- function(env_varname, env_var_list, default = env_is("LIBARROW_MINIMAL", "false")) {
+  # look in the environment first, but then use the env_var_list if nothing is found
+  env_var_list_value <- env_var_list[[env_varname]]
+  if (is.null(env_var_list_value)) {
+    env_var_list_value <- ""
+  }
+  env_value <- tolower(Sys.getenv(env_varname, env_var_list_value))
   if (identical(env_value, "off")) {
     # If e.g. ARROW_MIMALLOC=OFF explicitly, override default
     requested <- FALSE
@@ -839,8 +844,8 @@ is_feature_requested <- function(env_varname, default = env_is("LIBARROW_MINIMAL
 }
 
 with_cloud_support <- function(env_var_list) {
-  arrow_s3 <- is_feature_requested("ARROW_S3")
-  arrow_gcs <- is_feature_requested("ARROW_GCS")
+  arrow_s3 <- is_feature_requested("ARROW_S3", env_var_list)
+  arrow_gcs <- is_feature_requested("ARROW_GCS", env_var_list)
 
   if (arrow_s3 || arrow_gcs) {
     # User wants S3 or GCS support.
@@ -976,7 +981,7 @@ if (download_libarrow_ok && !is_release && !test_mode) {
 thirdparty_dependency_dir <- Sys.getenv("ARROW_THIRDPARTY_DEPENDENCY_DIR", "tools/thirdparty_dependencies")
 
 arrow_versioned <- paste0("arrow-", VERSION)
-# configure.win uses a different libarrow dir and and the zip is already nested
+# configure.win uses a different libarrow dir and the zip is already nested
 if (on_windows) {
   lib_dir <- "windows"
   dst_dir <- lib_dir
